static int has_data(int fd, int waitsecs)
{
    struct timeval tv = { .tv_sec = waitsecs, .tv_usec = 0 };
    fd_set rd;

    FD_ZERO(&rd); FD_SET(fd,&rd);
    return select(fd+1, &rd, NULL, NULL, &tv);
}

static int evtchn_ioctl(int evtchnd, int32_t request, void *data)
{
    struct evtchn_ioctl_msg req;
    struct evtchn_ioctl_msg rsp;
    int size = _IOC_SIZE(request);
    int rc;

    if (size > sizeof(req.data)) {
	fprintf(stderr, "%s: data too big: %d/%zd\n",
		__FUNCTION__, size, sizeof(req.data));
	return -1;
    }

    memset(&req, 0, sizeof(req));
    memcpy(req.data, data, size);
    req.ioctl = request;
    rc = write(evtchnd, &req, sizeof(req));
    if (rc != sizeof(req)) {
	fprintf(stderr, "%s: write: %d/%zd\n", __FUNCTION__, rc, sizeof(req));
	return -1;
    }

retry:
    if (!has_data(evtchnd, 5)) {
	fprintf(stderr, "%s: no reply within 5 sec\n", __FUNCTION__);
	return -1;
    }
    rc = read(evtchnd, &rsp, sizeof(rsp));
    if (rc != sizeof(rsp)) {
	fprintf(stderr, "%s: read: %d/%zd\n", __FUNCTION__, rc, sizeof(rsp));
	return -1;
    }
    if (req.ioctl != rsp.ioctl) {
	if (rsp.ioctl == EVTCHND_NOTIFY) {
	    if (libxc_trace)
		fprintf(stderr, "%s: got async notify, push back\n", __FUNCTION__);
	    write(evtchnd, &rsp, sizeof(rsp));
	} else if (libxc_fixme) {
	    fprintf(stderr, "%s: FIXME: unexpected reply\n", __FUNCTION__);
	}
	goto retry;
    }

    memcpy(data, rsp.data, size);
    if (-1 == rsp.retval)
	errno = rsp.error;
    return rsp.retval;
}

int xc_evtchn_open(void)
{
    struct sockaddr_in in;
    int rc, evtchnd;

    if (-1 == (evtchnd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP))) {
	perror("socket");
	rc = -1;
	goto out;
    }
    in.sin_family      = AF_INET;
    in.sin_port        = htons(EVTCHND_PORT);
    in.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    if (-1 == connect(evtchnd, (struct sockaddr*) &in, sizeof(in))) {
	perror("connect");
	close(evtchnd);
	rc = -1;
	goto out;
    }
    rc = evtchnd;

out:
    if (libxc_trace)
	fprintf(stderr, "libxc: %s -> handle %d\n",
		__FUNCTION__, rc);
    return rc;
}

int xc_evtchn_close(int evtchnd)
{
    close(evtchnd);
    return 0;
}

int xc_evtchn_fd(int evtchnd)
{
    return evtchnd;
}

evtchn_port_t xc_evtchn_bind_unbound_port(int evtchnd, int domid)
{
    struct ioctl_evtchn_bind_unbound_port io = {
	.remote_domain = domid,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_BIND_UNBOUND_PORT, &io);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, domid %d) -> %d\n",
		__FUNCTION__, evtchnd, domid, rc);
    return rc;
}

evtchn_port_t xc_evtchn_bind_interdomain(int evtchnd, int domid,
					 evtchn_port_t remote_port)
{
    struct ioctl_evtchn_bind_interdomain io = {
	.remote_domain = domid,
	.remote_port   = remote_port,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_BIND_INTERDOMAIN, &io);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, domid %d, rport %d) -> %d\n",
		__FUNCTION__, evtchnd, domid, remote_port, rc);
    return rc;
}

evtchn_port_t xc_evtchn_bind_virq(int evtchnd, unsigned int virq)
{
    struct ioctl_evtchn_bind_virq io = {
	.virq = virq,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_BIND_VIRQ, &io);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, virq %d) -> %d\n",
		__FUNCTION__, evtchnd, virq, rc);
    return rc;
}

int xc_evtchn_unbind(int evtchnd, evtchn_port_t port)
{
    struct ioctl_evtchn_unbind io = {
	.port = port,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_UNBIND, &io);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, port %d) -> %d\n",
		__FUNCTION__, evtchnd, port, rc);
    return rc;
}

int xc_evtchn_notify(int evtchnd, evtchn_port_t port)
{
    struct ioctl_evtchn_notify io = {
	.port = port,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_NOTIFY, &io);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, port %d) -> %d\n",
		__FUNCTION__, evtchnd, port, rc);
    return rc;
}

evtchn_port_t xc_evtchn_pending(int evtchnd)
{
    struct evtchn_ioctl_msg msg;
    struct evtchnd_port *n = (void*)(&msg.data);
    int rc;

    rc = read(evtchnd, &msg, sizeof(msg));
    if (rc != sizeof(msg)) {
	fprintf(stderr, "libxc: %s: read error (%d/%zd) %s\n",
		__FUNCTION__, rc, sizeof(msg), strerror(errno));
	return -1;
    }
    if (EVTCHND_NOTIFY != msg.ioctl) {
	fprintf(stderr, "libxc: %s: Huh? msg != notify?\n", __FUNCTION__);
	return -1;
    }
    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d) -> %d\n",
		__FUNCTION__, evtchnd, n->port);
    return n->port;
}

int xc_evtchn_unmask(int evtchnd, evtchn_port_t port)
{
    struct evtchn_ioctl_msg msg;
    struct evtchnd_port *n = (void*)(&msg.data);

    memset(&msg, 0, sizeof(msg));
    n->port = port;
    msg.ioctl = EVTCHND_UNMASK;
    write(evtchnd, &msg, sizeof(msg));

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, port %d)\n",
		__FUNCTION__, evtchnd, port);
    return 0;
}


/* ----------------------------------------------------------- */

int xenner_evtchnd_domid(int evtchnd, int domid);
int xenner_evtchnd_domid(int evtchnd, int domid)
{
    struct ioctl_evtchnd_domid io = {
	.domid = domid,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHND_DOMID, &io);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, domid %d) -> %d\n",
		__FUNCTION__, evtchnd, domid, rc);
    return rc;
}

