#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <xen/elfnote.h>
#include <xen/elfstructs.h>

#include "xenner.h"

/* ------------------------------------------------------------------ */

static const char *corepath[] = {
    VMCORE_PATH,
};

int vmcore_init(struct xenvm *xen)
{
    int i;

    for (i = 0; i < sizeof(corepath)/sizeof(corepath[0]); i++) {
	snprintf(xen->corefile, sizeof(xen->corefile), "%s/vmcore.%d",
		 corepath[i], xen->domid);
	xen->corefd = open(xen->corefile, O_CREAT | O_EXCL | O_RDWR, 0600);
	if (xen->corefd >= 0)
	    break;
	if (EEXIST == errno)
	    goto fail;
    }
    if (i == sizeof(corepath)/sizeof(corepath[0]))
	goto fail;

    /* create and map sparse file */
    xen->coresize = VMCORE_HEADER + 4096 * xen->pg_total;
    lseek(xen->corefd, xen->coresize -1, SEEK_SET);
    write(xen->corefd, "\0", 1);
    xen->coremap = mmap(NULL, xen->coresize, PROT_READ | PROT_WRITE,
			MAP_SHARED, xen->corefd, 0);
    if ((void*)-1 == xen->coremap) {
	fprintf(stderr, "mmap: %s\n", strerror(errno));
	return -1;
    }
    xen->memory = xen->coremap + VMCORE_HEADER;
    return 0;

fail:
    fprintf(stderr, "open %s: %s\n", xen->corefile, strerror(errno));
    return -1;
}

void vmcore_fini(struct xenvm *xen)
{
    munmap(xen->coremap, xen->coresize);
    unlink(xen->corefile);
}

/* ------------------------------------------------------------------ */

#define VMCORE_PHOFF  (PAGE_SIZE * 1)
#define VMCORE_SHOFF  (PAGE_SIZE * 2)

static Elf64_Phdr phdrs[] = {
    {
	/* physical memory */
	.p_type    = PT_LOAD,
	.p_flags   = PF_R | PF_W | PF_X,
	.p_offset  = VMCORE_HEADER,
	.p_paddr   = 0,
	.p_filesz  = 0, /* to be filled */
    }
};

static Elf64_Shdr shdrs[] = {
};

static Elf64_Ehdr ehdr = {
    .e_ident     = {
	[ EI_MAG0 ]       = ELFMAG0,
	[ EI_MAG1 ]       = ELFMAG1,
	[ EI_MAG2 ]       = ELFMAG2,
	[ EI_MAG3 ]       = ELFMAG3,
	[ EI_CLASS ]      = ELFCLASS64,
	[ EI_DATA ]       = ELFDATA2LSB,
 	[ EI_VERSION ]    = EV_CURRENT,
	[ EI_OSABI ]      = ELFOSABI_SYSV,
	[ EI_ABIVERSION ] = EV_CURRENT,
    },
    .e_type      = ET_CORE,
    .e_machine   = EM_X86_64,
    .e_version   = EV_CURRENT,
    .e_ehsize    = sizeof(Elf64_Ehdr),

    .e_phoff     = VMCORE_PHOFF,
    .e_phentsize = sizeof(phdrs[0]),
    .e_phnum     = sizeof(phdrs)/sizeof(phdrs[0]),

    .e_shoff     = VMCORE_SHOFF,
    .e_shentsize = sizeof(shdrs[0]),
    .e_shnum     = sizeof(shdrs)/sizeof(shdrs[0]),

#if 0
    .e_shstrndx  = ,
#endif
};

int vmcore_write_headers(struct xenvm *xen)
{
    lseek(xen->corefd, 0, SEEK_SET);
    write(xen->corefd, &ehdr, sizeof(ehdr));

    if (ehdr.e_phnum) {
	phdrs[0].p_filesz = xen->pg_total * PAGE_SIZE;

	lseek(xen->corefd, ehdr.e_phoff, SEEK_SET);
	write(xen->corefd, phdrs, sizeof(phdrs));
    }

    if (ehdr.e_shnum) {
	lseek(xen->corefd, ehdr.e_shoff, SEEK_SET);
	write(xen->corefd, shdrs, sizeof(shdrs));
    }
    return 0;
}
