#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <errno.h>

#include <xs.h>
#include <xen/io/xenbus.h>
#include <xen/io/xs_wire.h>

#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

static int xs_send_message(struct xenvm *xen, uint8_t *msg, int len)
{
    uint64_t paddr = frame_to_addr(xen->xenstore_pfn);
    struct xenstore_domain_interface *xs = guest_paddr_to_ptr(xen, paddr);
    XENSTORE_RING_IDX cons, prod, i;

    cons = xs->rsp_cons;
    prod = xs->rsp_prod;

    if (prod - cons > sizeof(xs->rsp))
	return -EIO;

    if (sizeof(xs->rsp) - (prod - cons) -1 < len)
	return -EAGAIN;

    /* FIXME: mb */

    for (i = 0; i < len; i++, prod++)
	xs->rsp[MASK_XENSTORE_IDX(prod)] = msg[i];

    /* FIXME: wmb */

    xs->rsp_prod += len;

    raise_event(xen, 0, xen->xenstore_event);
    return 0;
}

static void xs_print_message(struct xenvm *xen, char *prefix, uint8_t *msg, int len)
{
    struct xsd_sockmsg *hdr = (void*)msg;

    switch (hdr->type) {
    case XS_DIRECTORY:
	d0printf("%s: directory \"%s\"\n", prefix, msg+16);
	break;
    case XS_READ:
	d0printf("%s: read \"%s\"\n", prefix, msg+16);
	break;
    case XS_WATCH:
	d0printf("%s: watch \"%s\"\n", prefix, msg+16);
	break;
    case XS_TRANSACTION_START:
	d0printf("%s: transaction end\n", prefix);
	break;
    case XS_TRANSACTION_END:
	d0printf("%s: transaction end\n", prefix);
	break;
    case XS_ERROR:
	d0printf("%s: error \"%s\"\n", prefix, msg+16);
	break;
    default:
	d0printf("%s: unknown type %d\n", prefix, hdr->type);
    }
#if 0
    hexdump(prefix, msg, len);
#endif
}

static int xs_handle_message(struct xenvm *xen, uint8_t *msg, int len)
{
    uint8_t reply[32];
    struct xsd_sockmsg *hdr = (void*)msg;
    struct xsd_sockmsg *re  = (void*)reply;

    xs_print_message(xen, "xenbus << ", msg, len);

    memset(&reply, 0, sizeof(reply));
    re->type   = XS_ERROR;
    re->req_id = hdr->req_id;
    re->len    = 4;
    memcpy(reply + 16, "EIO", re->len);

    xs_print_message(xen, "xenbus >> ", reply, re->len + 16);
    xs_send_message(xen, reply, re->len + 16);
    return 0;
}

static int xenstore_event_internal(struct xenvm *xen)
{
    uint64_t paddr = frame_to_addr(xen->xenstore_pfn);
    struct xenstore_domain_interface *xs = guest_paddr_to_ptr(xen, paddr);
    XENSTORE_RING_IDX cons, prod, i;
    uint8_t buf[sizeof(xs->req)+1];
    struct xsd_sockmsg *hdr = (void*)buf;
    
    cons = xs->req_cons;
    prod = xs->req_prod;
    /* FIXME: mb */

    if (cons == prod || (prod - cons) > sizeof(xs->req))
	return 0;

    for (;;) {
	if (prod - cons < 16) {
	    d2printf("%s: msg head incomplete (%d/16)\n", __FUNCTION__,
		     prod - cons);
	    break;
	}
	for (i = 0; i < 16; i++, cons++)
	    buf[i] = xs->req[MASK_XENSTORE_IDX(cons)];
	hdr = (void*)buf;
	if (prod - cons < hdr->len) {
	    d2printf("%s: msg body incomplete (%d/%d)\n", __FUNCTION__,
		     prod - cons, hdr->len);
	    break;
	}
	for (; i < hdr->len + 16; i++, cons++)
	    buf[i] = xs->req[MASK_XENSTORE_IDX(cons)];
	buf[i] = 0;

	/* FIXME: mb */
	xs->req_cons = cons;

	xs_handle_message(xen, buf, i);
    }
    return 0;
}

/* ------------------------------------------------------------------ */

int xenstore_init(struct xenvm *xen)
{
    xen->xenstore = xs_daemon_open();
    if (NULL == xen->xenstore) {
	d0printf("%s: can't connect to %s (xenstored not running?)\n",
		 __FUNCTION__, xs_daemon_socket());
	return -1;
    }
    xs_introduce_domain(xen->xenstore,
			xen->domid,
			xen->xenstore_pfn + xen->mfn_guest,
			xen->xenstore_event);
    return 0;
}

void xenstore_fini(struct xenvm *xen)
{
    if (!xen->xenstore)
	return;
    xs_release_domain(xen->xenstore, xen->domid);
    xs_daemon_close(xen->xenstore);
}

int xenstore_event(struct xenvm *xen)
{
    if (!xen->xenstore) {
	xenstore_event_internal(xen);
    } else {
	evtchn_notify(xen, xen->xenstore_event);
    }
    return 0;
}
